/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Menu
Version:   1.0
ID:        ID_MENU
Status:    In Development
Category:  GUI
Date:      January 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2002-2004.  All rights reserved.
Keywords:  menu, menubar, list, option, select
Short:     The Menu class is used to construct menus from XML files.

###DESCRIPTION###

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/data/config.h>
#include <pandora/data/xml.h>
#include <pandora/files/file.h>
#include <pandora/graphics/all.h>
#include <pandora/input/joyport.h>
#include <pandora/input/keyboard.h>
#include <pandora/input/keymaptable.h>
#include <pandora/modules/filesystem.h>
#include <pandora/modules/pointer.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/render.h>
#include <pandora/system/all.h>
#include <pandora/functions/set.c>

#define VER_MENU 1.0

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_MENU, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2002-2004.  All rights reserved.",
   "February 2004",
   "Menu"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR MenuClass = 0;
static OBJECTPTR modStrings = NULL, modRender = NULL, modFileSystem = NULL, modPointer = NULL;
static struct PointerBase *PointerBase;
static struct StringsBase *StringsBase;
static struct RenderBase *RenderBase;
static struct FileSystemBase *FileSystemBase;
static OBJECTID glUserFocusID = NULL;
static UBYTE glLocale[4] = "eng";

/*****************************************************************************
** Key Translation Table
*/

struct Translation {
   LONG Code;
   STRING Name;
};

static struct Translation QualifierTable[] = {
   { KQ_LSHIFT,   "lshift" },
   { KQ_RSHIFT,   "rshift" },
   { KQ_CAPSLOCK, "capslock" },
   { KQ_LCONTROL, "lctrl" },
   { KQ_RCONTROL, "rctrl" },
   { KQ_LALT,     "lalt" },
   { KQ_RALT,     "ralt" },
   { KQ_LCOMMAND, "lcommand" },
   { KQ_RCOMMAND, "rcommand" },
   { KQ_NUMPAD,   "numpad" },
   /*** Pairs ***/
   { KQ_SHIFT,    "shift" },
   { KQ_COMMAND,  "command" },
   { KQ_ALT,      "alt" },
   { KQ_CONTROL,  "ctrl" },
   { KQ_CONTROL,  "control" },
   { NULL, NULL }
};

/****************************************************************************
** MenuItem structure.
*/

#define LEN_KEYSTRING 30

struct MenuItem {
   OBJECTID MenuID;        /* If the item refers to another menu, its ID must be indicated here */
   OBJECTID PictureID;
   LONG     Flags;         /* Special flags */
   LONG     YCoord;        /* Vertical coordinate inside the menu */
   LONG     Height;        /* Height of the item */
   STRING   Name;          /* Internal name of the menu item (for searching purposes) */
   STRING   Text;          /* Text to print in the menu item */
   STRING   Location;      /* Location of a menu definition file if this is an extension */
   STRING   ChildXML;      /* */
   STRING   Commands;      /* Command strings */
   LONG     Key;
   LONG     QualifierFlags;
   UBYTE    KeyString[LEN_KEYSTRING];
   struct   RGB Colour;     /* Font colour */
   struct   RGB Background; /* Background colour */
};

#define MENUITEM_DISABLED      0x00000001  /* Set if the menu item is disabled (ghosted) */
#define MENUITEM_BREAK         0x00000002  /* Menu item is acting as a break */
#define MENUITEM_EXTENSION     0x00000004  /* Menu item is an extension */
#define MENUITEM_CATEGORISE    0x00000008  /* Categorise the attached config file */
#define MENUITEM_FONTCOLOUR    0x00000010  /* Font colour is different to the default */
#define MENUITEM_FONTBOLD      0x00000020  /* Font is printed in bold */
#define MENUITEM_FONTITALICS   0x00000040  /* Font is printed in italics */
#define MENUITEM_BACKGROUND    0x00000080  /* Background of the menu item should be drawn in a special colour */
#define MENUITEM_NOKEYRESPONSE 0x00000100  /* Display the key, but do not respond to its equivalent key press */
#define MENUITEM_KEYREPEAT     0x00000200  /* Respond to repeated key presses */
#define MENUITEM_SORT          0X00000400  /* Sort all content under the menuitem (applies only to extensions) */

/****************************************************************************
** Menu class.
*/

#define MAX_ITEMS 100
#define LEN_NODE 32

struct Menu {
   OBJECT_HEADER
   struct Font *Font;       /* The font being used in the menu */
   OBJECTID DrawableID;     /* Drawable for this menu */
   OBJECTID TargetID;       /* Target for the drawable */
   OBJECTID ParentID;       /* Parent menu if this is a child of a master menu */
   OBJECTID RelativeID;     /* Relative drawable that should be used for coordinate origins */
   OBJECTID KeyMonitorID;   /* Drawable to monitor for key presses */
   OBJECTID KeyboardID;     /* Keyboard to use for listening to key presses */
   OBJECTID SelectionObjectID;
   STRING   Location;       /* Location of the XML or config menu definition file */
   LONG     TotalItems;     /* Total number of items in the menu */
   LONG     Flags;          /* */
   LONG     VSpacing;       /* Amount of spacing between each menu item */
   LONG     BreakHeight;    /* The amount of height to give to menu-break graphics */
   LONG     Width;
   LONG     LeftMargin;     /* Left hand margin inside the menu */
   LONG     RightMargin;    /* Right hand margin inside the menu */
   LONG     TopMargin;      /* Top margin inside the menu */
   LONG     BottomMargin;   /* Bottom margin inside the menu */
   LONG     Duration;       /* The length of time allocated to special effects like fading.  Value in milliseconds */
   LONG     AutoExpand;     /* Auto-expand timeout setting */
   LONG     HighlightLM;    /* Highlight rectangle left margin */
   LONG     HighlightRM;    /* Highlight rectangle right margin */
   LONG     ItemHeight;     /* Minimum allowable height for text based menu items */
   LONG     ImageSize;      /* Size of the image column.  Icons will be generated to fit this size */

   LONG     Highlight;      /* Index of the item that is currently highlighted */
   LONG     ImageGap;       /* Gap between the image column and text */
   LONG     KeyGap;         /* Gap between the text and key columns */
   LONG     ExtensionGap;   /* Gap between the extension column and the text or key column */
   LONG     TextWidth;      /* Width of the widest text string */
   LONG     KeyWidth;       /* Width of the widest key string */

   PRIVATE_FIELDS
   struct MenuItem * Items[MAX_ITEMS];    /* Chain of items listed in the menu */
   UBYTE     Language[4];          /* 3-character language code, null-terminated */
   UBYTE     LanguageDir[32];      /* Directory to use for language files */
   OBJECTPTR prvXML;
   OBJECTPTR Translation;          /* Translation data (config object) */
   DOUBLE    XCoord;
   DOUBLE    YCoord;
   OBJECTID  RootID;               /* Reference to the top-most menu */
   OBJECTID  CurrentMenuID;        /* Currently active sub-menu */
   BYTE      SelectionField[32];
   LONG      prvBitsPerPixel;
   LONG      TotalArgs;            /* Total number of arguments in the LocalArgs table */
   STRING    *LocalArgs;           /* Argument table (created via calls to SetUnlistedField()) */
   WORD      KeyboardChannel;
   STRING    prvBaseScript;
   STRING    prvItemScript;
   STRING    prvBreakScript;
   struct    RGB prvFontColour;
   struct    RGB prvFontHighlight;
   struct    RGB prvHighlight;
   struct    RGB prvHighlightBorder;
   UBYTE     prvNode[LEN_NODE];
   WORD      ParentIndex;
   UBYTE     prvFade;
   ULONG     FadeTimer;
   WORD      MouseOver;
   LONG      VWhiteSpace;
};

enum {
   MENUFADE_FADEIN=1,
   MENUFADE_FADEOUT
};

#define MND_DOWN 1
#define MND_UP   2

/*****************************************************************************
** Structural definition.
*/

static ERROR GET_Location(struct Menu *, STRING *);
static ERROR GET_Node(struct Menu *, STRING *);

static ERROR SET_FontColour(struct Menu *, STRING);
static ERROR SET_FontHighlight(struct Menu *, STRING);
static ERROR SET_Highlight(struct Menu *, STRING);
static ERROR SET_HighlightBorder(struct Menu *, STRING);
static ERROR SET_Location(struct Menu *, STRING);
static ERROR SET_MenuFile(struct Menu *, STRING);
static ERROR SET_Node(struct Menu *, STRING);
static ERROR SET_SelectionField(struct Menu *, STRING);
static ERROR SET_Width(struct Menu *, LONG *);
static ERROR SET_X(struct Menu *, LONG *);
static ERROR SET_Y(struct Menu *, LONG *);

static struct FieldDef MenuFlags[] = {
   { "SHOWIMAGES",  MNF_SHOWIMAGES },
   { "SHOWKEYS",    MNF_SHOWKEYS },
   { "EXTCOLUMN",   MNF_EXTCOLUMN },
   { "POPUP",       MNF_POPUP },
   { "FADEIN",      MNF_FADEIN },
   { "FADEOUT",     MNF_FADEOUT },
   { "NOHIDE",      MNF_NOHIDE },
   { "IGNOREFOCUS", MNF_IGNOREFOCUS },
   /*** Synonyms ***/
   { "FADE",        MNF_FADEIN|MNF_FADEOUT },
   { NULL, NULL }
};

static struct FieldArray MenuFields[] = {
   { "Font",            0, FDF_CHILD|FDF_R,      0, NULL, NULL },
   { "Drawable",        0, FDF_OBJECTID|FDF_R,   0, NULL, NULL },
   { "Target",          0, FDF_OBJECTID|FDF_RI,  0, NULL, NULL },
   { "Parent",          0, FDF_OBJECTID|FDF_RI,  0, NULL, NULL },
   { "Relative",        0, FDF_OBJECTID|FDF_RW,  0, NULL, NULL },
   { "KeyMonitor",      0, FDF_OBJECTID|FDF_RW,  0, NULL, NULL },
   { "Keyboard",        0, FDF_OBJECTID|FDF_RI,  0, NULL, NULL },
   { "SelectionObject", 0, FDF_OBJECTID|FDF_RW, 0, NULL, NULL },
   { "Location",        0, FDF_STRING|FDF_RW,    0, NULL, SET_Location },
   { "TotalItems",      0, FDF_LONG|FDF_R,       0, NULL, NULL },
   { "Flags",           0, FDF_LONGFLAGS|FDF_RW, (LONG)&MenuFlags, NULL, NULL },
   { "VSpacing",        0, FDF_LONG|FDF_RI,      0, NULL, NULL },
   { "BreakHeight",     0, FDF_LONG|FDF_R,       0, NULL, NULL },
   { "Width",           0, FDF_LONG|FDF_RW,      0, NULL, SET_Width },
   { "LeftMargin",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "RightMargin",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "TopMargin",       0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "BottomMargin",    0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Duration",        0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "AutoExpand",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "HighlightLM",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "HighlightRM",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ItemHeight",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ImageSize",       0, FDF_LONG|FDF_RW,      0, NULL, NULL },

   /*** Virtual fields ***/
   { "FontColour",      0, FDF_STRING|FDF_W,    0, NULL,         SET_FontColour },
   { "FontHighlight",   0, FDF_STRING|FDF_W,    0, NULL,         SET_FontHighlight },
   { "Highlight",       0, FDF_STRING|FDF_W,    0, NULL,         SET_Highlight },
   { "HighlightBorder", 0, FDF_STRING|FDF_W,    0, NULL,         SET_HighlightBorder },
   { "MenuFile",        0, FDF_STRING|FDF_INIT, 0, NULL,         SET_MenuFile },
   { "Node",            0, FDF_STRING|FDF_RW,   0, GET_Node,     SET_Node },
   { "SelectionField",  0, FDF_STRING|FDF_W,    0, NULL,         SET_SelectionField },
   { "Src",             0, FDF_STRING|FDF_RW,   0, GET_Location, SET_Location },
   { "X",               0, FDF_LONG|FDF_RW,     0, NULL,         SET_X },
   { "XCoord",          0, FDF_LONG|FDF_RW,     0, NULL,         SET_X },
   { "Y",               0, FDF_LONG|FDF_RW,     0, NULL,         SET_Y },
   { "YCoord",          0, FDF_LONG|FDF_RW,     0, NULL,         SET_Y },
   END_FIELD
};

/*****************************************************************************
** Menu actions.
*/

static ERROR MENU_ActionNotify(struct Menu *, struct acActionNotify *);
static ERROR MENU_Activate(struct Menu *, APTR);
static ERROR MENU_DataChannel(struct Menu *, struct acDataChannel *);
static ERROR MENU_Focus(struct Menu *, APTR);
static ERROR MENU_Free(struct Menu *, APTR);
static ERROR MENU_Hide(struct Menu *, APTR);
static ERROR MENU_Init(struct Menu *, APTR);
static ERROR MENU_LostFocus(struct Menu *, APTR);
static ERROR MENU_NewObject(struct Menu *, APTR);
static ERROR MENU_Refresh(struct Menu *, APTR);
static ERROR MENU_SetUnlistedField(struct Menu *, struct acSetUnlistedField *);
static ERROR MENU_Show(struct Menu *, APTR);
static ERROR MENU_Timer(struct Menu *, struct acTimer *);

static struct ActionArray MenuActions[] = {
   { AC_ActionNotify,     MENU_ActionNotify },
   { AC_Activate,         MENU_Activate },
   { AC_DataChannel,      MENU_DataChannel },
   { AC_Focus,            MENU_Focus },
   { AC_Free,             MENU_Free },
   { AC_Hide,             MENU_Hide },
   { AC_Init,             MENU_Init },
   { AC_LostFocus,        MENU_LostFocus },
   { AC_NewObject,        MENU_NewObject },
   { AC_Refresh,          MENU_Refresh },
   { AC_SetUnlistedField, MENU_SetUnlistedField },
   { AC_Show,             MENU_Show },
   { AC_Timer,            MENU_Timer },
   { NULL, NULL }
};

/*****************************************************************************
** Menu methods.
*/

#define MT_ExecuteMenuItem -1

struct mtExecuteMenuItem {
   LONG Index;
};

static ERROR MENU_ExecuteItem(struct Menu *, struct mtExecuteMenuItem *);

static struct MethodArray MenuMethods[] = {
   { MT_ExecuteMenuItem, MENU_ExecuteItem, "ExecuteItem", NULL, NULL },
   { NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Function prototypes.
*/

static ERROR AddXMLItem(struct Menu *, OBJECTPTR, struct XMLTag *);
static ERROR CalculateMenuSize(struct Menu *);
static ERROR CreateMenu(struct Menu *);
static void DrawMenu(struct Menu *, struct Render *, struct Bitmap *);
static void EnsureVisibility(struct Menu *);
static ERROR ExecuteItem(struct Menu *, LONG, LONG);
static ERROR GenerateMenuFile(struct Menu *, struct Menu *, LONG);
static WORD GetItemHeight(struct Menu *);
static void InitKeyboard(struct Menu *);
static ERROR LoadIcon(struct Menu *, STRING, OBJECTID *);
static ERROR WriteFile(OBJECTPTR, STRING);
static WORD ReadByte(STRING);
static ERROR ReadColour(STRING, LONG *, LONG *, LONG *);
static ERROR HighlightItem(struct Menu *, LONG);
static void ParseXMLTag(struct Menu *, OBJECTPTR, struct XMLTag *);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   OBJECTPTR config;

   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
         FID_Name|TSTRING,   "strings",
         FID_Version|TFLOAT, MODVERSION_STRINGS,
         TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modRender, NULL,
         FID_Name|TSTRING,   "render",
         FID_Version|TFLOAT, MODVERSION_RENDER,
         TAGEND) IS ERR_Okay) {
      if (GetField(modRender, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modPointer, NULL,
         FID_Name|TSTRING,   "pointer",
         FID_Version|TFLOAT, MODVERSION_POINTER,
         TAGEND) IS ERR_Okay) {
      if (GetField(modPointer, FID_ModBase, FT_POINTER, &PointerBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
         FID_Name|TSTRING,   "filesystem",
         FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
         TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   /*** Get the user's locale for translation purposes ***/

   if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
      FID_Location|TSTRING, "user:config/locale.cfg",
      TAGEND) IS ERR_Okay) {
      struct mtReadConfig read;
      read.Section = "LOCALE";
      read.Item    = "Language";
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
         if (StrLength(read.Result) IS 3) {
            StrCopy(read.Result, glLocale, sizeof(glLocale));
            StrLower(glLocale);
         }
      }
      Action(AC_Free, config, NULL);
   }

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&MenuClass, NULL,
      FID_BaseClassID|TLONG, ID_MENU,
      FID_Version|TFLOAT,    VER_MENU,
      FID_Name|TSTRING,      "Menu",
      FID_Category|TLONG,    CCF_GUI,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN,
      FID_Actions|TPTR,      MenuActions,
      FID_Methods|TPTR,      MenuMethods,
      FID_Fields|TPTR,       MenuFields,
      FID_Size|TLONG,        sizeof(struct Menu),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (modStrings)    { acFree(modStrings);    modStrings = 0; }
   if (modRender)     { acFree(modRender);     modRender = 0; }
   if (modPointer)    { acFree(modPointer);    modPointer = 0; }
   if (modFileSystem) { acFree(modFileSystem); modFileSystem = 0; }
   if (MenuClass)     { acFree(MenuClass);     MenuClass = 0; }
   return(ERR_Okay);
}

/*****************************************************************************
** Menu: ActionNotify
*/

static ERROR MENU_ActionNotify(struct Menu *Self, struct acActionNotify *NotifyArgs)
{
   LONG action;

   if (!NotifyArgs) return(ERR_Args);
   if (NotifyArgs->Error != ERR_Okay) return(ERR_Okay);

   action = NotifyArgs->ActionID;

   if (action IS AC_Focus) {
      InitKeyboard(Self);

      if (NotifyArgs->ObjectID IS Self->DrawableID) {
         glUserFocusID = Self->Head.UniqueID;
      }
   }
   else if (action IS AC_Hide) {
      if (Self->DrawableID IS NotifyArgs->ObjectID) {
         /*** Our drawable has been hidden ***/

         if (glUserFocusID IS Self->Head.UniqueID) glUserFocusID = NULL;
         Self->Highlight = -1;

         /* If we are a child menu, commit suicide when we are no longer
         ** on display.  The parent can recreate us when it wants to.
         */

/* Uncommented because killing the child menu also results in removing
   support for its key presses (MNF_SHOWKEYS).

         if (Self->ParentID) {
            acFree(Self);
            return(ERR_Okay);
         }
*/
      }
      else {
         /*** The parent menu has disappeared ***/
         ActionMsg(AC_Hide, Self->DrawableID, NULL);
         Self->CurrentMenuID = NULL;
      }

      if ((NotifyArgs->ObjectID IS Self->KeyMonitorID) OR (!(Self->Flags & MNF_SHOWKEYS))) {
         Action(AC_LostFocus, Self, NULL);
      }

      /*** This code just ensures that the opacity is set to zero (for when the menu has to fade in next time) ***/

      Self->prvFade = 0;
      if (Self->Flags & MNF_FADEIN) {
         struct mtSetRenderOpacity opacity;
         opacity.Value = 0;
         opacity.Adjustment = 0;
         ActionMsg(MT_SetRenderOpacity, Self->DrawableID, &opacity);
      }
   }
   else if (action IS AC_LostFocus) {
      if (NotifyArgs->ObjectID IS Self->DrawableID) {
         if (glUserFocusID IS Self->Head.UniqueID) glUserFocusID = NULL;
      }

      if ((NotifyArgs->ObjectID IS Self->KeyMonitorID) OR (!(Self->Flags & MNF_SHOWKEYS))) {
         Action(AC_LostFocus, Self, NULL);
      }

      if (!(Self->Flags & MNF_IGNOREFOCUS)) {
         if ((NotifyArgs->ObjectID IS Self->RelativeID) OR ((NotifyArgs->ObjectID IS Self->DrawableID) AND (!Self->ParentID))) {
            /* (Hide if I am the root menu and I lost the focus) OR
            ** (My relative drawable lost the focus).
            */

            Action(AC_Hide, Self, NULL);

            /*** Hide sub-menus ***/

            if (Self->CurrentMenuID) {
               ActionMsg(AC_Hide, Self->CurrentMenuID, NULL);
               Self->CurrentMenuID = NULL;
            }
         }
      }
   }
   else if (action IS AC_Show) {
      if (Self->Flags & MNF_FADEIN) {
         Self->prvFade = MENUFADE_FADEIN;
         Self->FadeTimer = 0;
         SubscribeTimer(Self->Head.UniqueID, 20);
      }
   }
   else if (action IS AC_UserClick) {
      struct acUserClick *click;
      LONG y, i;

      click = (struct acUserClick *)NotifyArgs->Args;

      if ((NotifyArgs->ObjectID IS Self->DrawableID) AND (click->ButtonFlags & JD_LMB)) {
         y = Self->TopMargin;
         for (i=0; i < Self->TotalItems; i++) {
            if (!(Self->Items[i]->Flags & MENUITEM_BREAK)) {
               if ((click->YCoord >= y) AND (click->YCoord < y + Self->Items[i]->Height)) {
                  ExecuteItem(Self, i, FALSE);
                  return(ERR_Okay);
               }
            }

            y += Self->Items[i]->Height;
         }
      }
   }
   else if (action IS AC_UserMovement) {
      struct acUserMovement *move;
      LONG y, i;

      move = (struct acUserMovement *)NotifyArgs->Args;

      if (move->OverObject IS Self->DrawableID) {
         Self->MouseOver = Self->AutoExpand;
         SubscribeTimer(Self->Head.UniqueID, -20);

         y = Self->TopMargin;
         for (i=0; i < Self->TotalItems; i++) {
            if (!(Self->Items[i]->Flags & MENUITEM_BREAK)) {
               if ((move->YCoord >= y) AND (move->YCoord < y + Self->Items[i]->Height)) {
                  if (Self->Highlight != i) {
                     HighlightItem(Self, i);
                  }
                  return(ERR_Okay);
               }
            }

            y += Self->Items[i]->Height;
         }
      }

      /* Remove existing menu highlighting if the cursor is no longer
      ** positioned over a highlight-able item.
      */

      if (Self->Highlight != -1) HighlightItem(Self, -1);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Menu: Activate
*/

static ERROR MENU_Activate(struct Menu *Self, APTR Void)
{
   struct RenderInfo drawable_info;

   DPrintF("~Activate()","[Menu:%d]", Self->Head.UniqueID);

   if (Self->prvFade) {
      /*** Do not interfere with fading menus ***/
      DPrintF("Activate:","Menu is currently fading.");
   }
   else if (!(Self->Flags & MNF_NOHIDE)) {
      /*** If the menu is already visible, hide it ***/

      if (GetRenderInfo(Self->DrawableID, &drawable_info, sizeof(drawable_info)) IS ERR_Okay) {
         if (drawable_info.Flags & RNF_VISIBLE) {
            Action(AC_Hide, Self, NULL);
         }
         else Action(AC_Show, Self, NULL);
      }
   }
   else Action(AC_Show, Self, NULL);

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Menu: DataChannel
*/

static ERROR MENU_DataChannel(struct Menu *Self, struct acDataChannel *Args)
{
   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version IS 1)) {
      struct RenderInfo info;
      struct dcKeyEntry *key;
      LONG i, j;

      if ((key = (struct dcKeyEntry *)Args->Buffer)) {
         for (i=0; i < Args->TotalEntries; i++) {

            if (key[i].Flags & KQ_PRESSED) {
               if (key[i].Value IS K_DOWN) {
                  if (glUserFocusID IS Self->Head.UniqueID) {
                     j = Self->Highlight + 1;
                     while ((j < Self->TotalItems) AND (Self->Items[j]->Flags & MENUITEM_BREAK)) j++;
                     if (j >= Self->TotalItems) j = Self->TotalItems - 1;
                     HighlightItem(Self, j);
                  }
               }
               else if (key[i].Value IS K_UP) {
                  if (glUserFocusID IS Self->Head.UniqueID) {
                     j = Self->Highlight - 1;
                     while ((j > 0) AND (Self->Items[j]->Flags & MENUITEM_BREAK)) j--;
                     if (j < 0) j = 0;
                     HighlightItem(Self, j);
                  }
               }
               else if (key[i].Value IS K_LEFT) {
                  struct Menu *menu;
                  if ((glUserFocusID IS Self->Head.UniqueID)) {
                     if (Self->ParentID) {
                        if ((Self->Highlight >=0) AND (Self->Highlight < Self->TotalItems)) {
                           HighlightItem(Self, -1);
                           Action(AC_Hide, Self, NULL);
                           if (AccessObject(Self->ParentID, 4000, (OBJECTPTR *)&menu) IS ERR_Okay) {
                              glUserFocusID = Self->ParentID;
                              HighlightItem(menu, Self->ParentIndex);
                              acFocus(menu);
                              ReleaseObject(menu);
                           }
                        }
                     }
                     else if (Self->RelativeID) ActionMsg(AC_Focus, Self->RelativeID, NULL);
                     else if (GetRenderInfo(Self->DrawableID, &info, sizeof(info)) IS ERR_Okay) {
                        if (info.ParentID) ActionMsg(AC_Focus, info.ParentID, NULL);
                     }
                  }
               }
               else if (key[i].Value IS K_RIGHT) {
                  if (glUserFocusID IS Self->Head.UniqueID) {
                     if ((Self->Highlight >=0) AND (Self->Highlight < Self->TotalItems)) {
                        if (Self->Items[Self->Highlight]->Flags & MENUITEM_EXTENSION) {
                           ExecuteItem(Self, Self->Highlight, TRUE);
                        }
                     }
                  }
               }
               else if (key[i].Value IS K_ESCAPE) {
                  if (glUserFocusID IS Self->Head.UniqueID) {
                     if (Self->RelativeID) ActionMsg(AC_Focus, Self->RelativeID, NULL);
                     else if (GetRenderInfo(Self->DrawableID, &info, sizeof(info)) IS ERR_Okay) {
                        if (info.ParentID) ActionMsg(AC_Focus, info.ParentID, NULL);
                     }
                  }
               }
               else if ((key[i].Value IS K_ENTER) OR (key[i].Value IS K_SPACE)) {
                  if (glUserFocusID IS Self->Head.UniqueID) {
                     if ((Self->Highlight >=0) AND (Self->Highlight < Self->TotalItems)) {
                        ExecuteItem(Self, Self->Highlight, TRUE);
                     }
                  }
               }
               else for (j=0; j < Self->TotalItems; j++) {

                  if (key[i].Value IS Self->Items[j]->Key) {
                     if (Self->Items[j]->Flags & MENUITEM_NOKEYRESPONSE) break;

                     if (!(Self->Items[j]->Flags & MENUITEM_KEYREPEAT)) {
                        if (key[i].Flags & KQ_REPEAT) break;
                     }

                     if (Self->Items[j]->QualifierFlags) {
                        if (Self->Items[j]->QualifierFlags & (KQ_LCONTROL|KQ_RCONTROL)) {
                           if (!((key[i].Flags & (KQ_LCONTROL|KQ_RCONTROL)) & Self->Items[j]->QualifierFlags)) continue;
                        }

                        if (Self->Items[j]->QualifierFlags & (KQ_LSHIFT|KQ_RSHIFT)) {
                           if (!((key[i].Flags & (KQ_LSHIFT|KQ_RSHIFT)) & Self->Items[j]->QualifierFlags)) continue;
                        }

                        if (Self->Items[j]->QualifierFlags & (KQ_LALT|KQ_RALT)) {
                           if (!((key[i].Flags & (KQ_LALT|KQ_RALT)) & Self->Items[j]->QualifierFlags)) continue;
                        }

                        if (Self->Items[j]->QualifierFlags & (KQ_LCOMMAND|KQ_RCOMMAND)) {
                           if (!((key[i].Flags & (KQ_LCOMMAND|KQ_RCOMMAND)) & Self->Items[j]->QualifierFlags)) continue;
                        }
                     }

                     ExecuteItem(Self, j, FALSE);
                  }
               }
            }
         }
      }
   }
   else if (Args->DataType IS DATA_XML) {
      OBJECTPTR xml;
      struct XMLTag **xmllist, *tag;

      if (CreateObject(ID_XML, NF_CHILD, &xml, NULL,
         FID_Statement|TSTRING, Args->Buffer,
         NULL) IS ERR_Okay) {

         if (GetFields(xml, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
            for (tag=xmllist[0]; tag; tag=tag->Next) {
               ParseXMLTag(Self, xml, tag);
            }
         }

         Action(AC_Free, xml, NULL);

         /*** Recalculate the menu size ***/

         CalculateMenuSize(Self);

         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_DataChannel, ERR_CreateObject));
   }
   else DPrintF("DataChannel:","[Menu] Unrecognised data type %d, version %d.", Args->DataType, Args->Version);

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  ExecuteItem
Short: Executes the currently selected menu item.

###INPUT###
LONG/Index: The index of the item to be executed.

###END###

*****************************************************************************/

static ERROR MENU_ExecuteItem(struct Menu *Self, struct mtExecuteMenuItem *Args)
{
   if (!Args) return(ObjectError(Self, MT_ExecuteMenuItem, ERR_Args));


   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Sends the focus to the menu and activates keyboard monitoring.
###END###

*****************************************************************************/

static ERROR MENU_Focus(struct Menu *Self, APTR Void)
{
   return(ActionMsg(AC_Focus, Self->DrawableID, NULL));
}

/*****************************************************************************
** Menu: Free
*/

static ERROR MENU_Free(struct Menu *Self, APTR Void)
{
   OBJECTPTR object;
   LONG i;

   if (Self->Translation) {
      Action(AC_Free, Self->Translation, NULL);
      Self->Translation = NULL;
   }

   if (Self->LocalArgs) {
      for (i=0; i < Self->TotalArgs; i++) {
         if (Self->LocalArgs[i]) FreeMemory(Self->LocalArgs[i]);
      }
      FreeMemory(Self->LocalArgs);
      Self->LocalArgs = NULL;
   }

   for (i=0; i < Self->TotalItems; i++) {
      if (Self->Items[i]->PictureID) ActionMsg(AC_Free, Self->Items[i]->PictureID, NULL);
      if (Self->Items[i]->MenuID)    ActionMsg(AC_Free, Self->Items[i]->MenuID, NULL);
      if (Self->Items[i]->Name)      FreeMemory(Self->Items[i]->Name);
      if (Self->Items[i]->Text)      FreeMemory(Self->Items[i]->Text);
      if (Self->Items[i]->Location)  FreeMemory(Self->Items[i]->Location);
      if (Self->Items[i]->ChildXML)  FreeMemory(Self->Items[i]->ChildXML);
      if (Self->Items[i]->Commands)  FreeMemory(Self->Items[i]->Commands);

      FreeMemory(Self->Items[i]);
      Self->Items[i] = NULL;
   }

   if (Self->prvBaseScript) {
      FreeMemory(Self->prvBaseScript);
      Self->prvBaseScript = NULL;
   }

   if (Self->prvBreakScript) {
      FreeMemory(Self->prvBreakScript);
      Self->prvBreakScript = NULL;
   }

   if (Self->prvItemScript) {
      FreeMemory(Self->prvItemScript);
      Self->prvItemScript = NULL;
   }

   if (Self->Location) {
      FreeMemory(Self->Location);
      Self->Location = NULL;
   }

   if (Self->Font) {
      Action(AC_Free, Self->Font, NULL);
      Self->Font = NULL;
   }

   if (Self->prvXML) {
      Action(AC_Free, Self->prvXML, NULL);
      Self->prvXML = NULL;
   }

   if (Self->KeyboardChannel IS TRUE) {
      if (AccessObject(Self->KeyboardID, 4000, &object) IS ERR_Okay) {
         UnsubscribeChannel(object, Self->Head.UniqueID);
         ReleaseObject(object);
      }
   }

   if (Self->DrawableID) {
      if (AccessObject(Self->DrawableID, 4000, &object) IS ERR_Okay) {
         UnsubscribeAction(object, NULL, GetUniqueID(Self));
         ActionTags(MT_SetRenderCallback, object, Self, NULL);
         Action(AC_Free, object, NULL);
         ReleaseObject(object);
      }
      Self->DrawableID = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name: Hide
Short: Hides the menu and open sub-menus.
###END###

*****************************************************************************/

static ERROR MENU_Hide(struct Menu *Self, APTR Void)
{
   DPrintF("~Hide()","[Menu:%d]", Self->Head.UniqueID);

   if (Self->Flags & MNF_FADEOUT) {
      if (Self->prvFade) {
         Self->prvFade = MENUFADE_FADEOUT; /* A fade is already active - simply ensure that we use the fadeout routine */
         SubscribeTimer(Self->Head.UniqueID, 20);
      }
      else {
         Self->prvFade = MENUFADE_FADEOUT;
         Self->FadeTimer = 0;
         SubscribeTimer(Self->Head.UniqueID, 20);
      }
   }
   else {
      ActionMsg(AC_Hide, Self->DrawableID, NULL);
   }

   if (Self->CurrentMenuID) {
      /*** Hide any sub-menus ***/
      DPrintF("Hide:","Hiding sub-menu #%d.", Self->CurrentMenuID);
      ActionMsg(AC_Hide, Self->CurrentMenuID, NULL);
      Self->CurrentMenuID = NULL;
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Menu: Init
*/

static ERROR MENU_Init(struct Menu *Self, APTR Void)
{
   struct RenderInfo info;
   struct Menu *parent;
   OBJECTPTR object;
   OBJECTID ownerid;
   CLASSID class_id;
   ERROR error;

   if ((!Self->KeyMonitorID) AND (Self->Flags & MNF_SHOWKEYS)) {
      DPrintF("Init:","[Menu] Disabling the SHOWKEYS flag because the KeyMonitor field is not set.");
      Self->Flags &= ~MNF_SHOWKEYS;
   }

   if (!Self->KeyboardID) {
      if (FastFindObject("SystemKeyboard", NULL, &Self->KeyboardID, 1, NULL) != ERR_Okay) {
         DPrintF("@Init:","[Menu] Could not find any SystemKeyboard objects.");
         return(ERR_Failed);
      }
   }

   if (Self->HighlightLM IS -1) Self->HighlightLM = Self->LeftMargin;
   if (Self->HighlightRM IS -1) Self->HighlightRM = Self->RightMargin;

   /*** Create a font object for drawing the menu text ***/

   Self->Font->Flags |= FTF_DIRECTXY;
   if (Action(AC_Init, Self->Font, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Init));
   }

   if (Action(AC_Init, Self->prvXML, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Init));
   }

   /* If this is a child menu then subscribe to the hide action
   ** of our parent's drawable.
   */

   if (Self->ParentID) {
      if (AccessObject(Self->ParentID, 4000, (OBJECTPTR *)&parent) IS ERR_Okay) {
         if (parent->DrawableID) {
            if (AccessObject(parent->DrawableID, 4000, &object) IS ERR_Okay) {
               SubscribeActionTags(object, Self, AC_Hide, TAGEND);
               ReleaseObject(object);
            }
            else ObjectError(Self, ERH_Init, ERR_AccessObject);
         }
         ReleaseObject(parent);
      }
      else return(ObjectError(Self, ERH_Init, ERR_AccessObject));
   }

   /*** If we have no parent, we are the root menu ***/

   if (!Self->ParentID) Self->RootID = Self->Head.UniqueID;

   /*** If no target was given, set the target to the top-most render object ***/

   if (!Self->TargetID) {
      ownerid = GetOwner(Self);
      while (ownerid) {
         class_id = GetClassID(ownerid);
         if (class_id IS ID_RENDER) {
            Self->TargetID = ownerid;
            if (GetRenderInfo(ownerid, &info, sizeof(info)) IS ERR_Okay) {
               /*** Stop searching if we found a host drawable (e.g. the desktop) ***/
               if (info.Flags & RNF_HOST) break;
               ownerid = info.ParentID;
            }
            else ownerid = GetOwnerID(ownerid);
         }
         else if (class_id IS ID_WINDOW) {
            if (AccessObject(ownerid, 5000, &object) IS ERR_Okay) {
               GetField(object, FID_Drawable, FT_LONG, &ownerid);
               ReleaseObject(object);
            }
            else ownerid = GetOwnerID(ownerid);
         }
         else ownerid = GetOwnerID(ownerid);
      }
      if (!Self->TargetID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));

      DPrintF("Init:","[Menu] Target search found drawable #%d.", Self->TargetID);
   }

   /*** Create the menu ***/

   if ((error = CreateMenu(Self)) IS ERR_Okay) {
      if (Self->RelativeID) {
         if (AccessObject(Self->RelativeID, 4000, &object) IS ERR_Okay) {
            SubscribeAction(object, AC_LostFocus, Self);
            ReleaseObject(object);
         }
      }

      if (AccessObject(Self->DrawableID, 4000, &object) IS ERR_Okay) {
         ActionTags(MT_SetRenderCallback, object, Self, &DrawMenu);
         SubscribeActionTags(object, Self, AC_Show, AC_Hide, AC_Focus, AC_LostFocus, AC_UserClick, AC_UserMovement, TAGEND);
         GetField(object, FID_BottomMargin, FT_LONG, &Self->VWhiteSpace);
         ReleaseObject(object);
      }
      else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

      /* The KeyMonitor field determines what drawable object we should listen to for focus
      ** actions in determining when we should subscribe to the keyboard.  It often
      ** refers to menu-bars or windows, but for pop-up menus we subscribe to our
      ** own drawable so that we can at least support the cursor keys for highlighting.
      */

      if (!Self->KeyMonitorID) Self->KeyMonitorID = Self->DrawableID;

      if (AccessObject(Self->KeyMonitorID, 4000, &object) IS ERR_Okay) {
         if (Self->KeyMonitorID IS Self->DrawableID) {
            /* Do nothing (drawable is already subscribed to Show/Hide actions) */
         }
         else SubscribeActionTags(object, Self, AC_Focus, AC_LostFocus, TAGEND);

         /* Test the key-monitored object to see if it already has the focus and
         ** if so, activate keyboard monitoring.
         */

         if (object->ClassID IS ID_RENDER) {
            if (((struct Render *)object)->Flags & RNF_HASFOCUS) InitKeyboard(Self);
         }

         ReleaseObject(object);
      }

      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_Init, error));
}

/*****************************************************************************

###ACTION###
Name:  LostFocus
Short: Removes the focus from the menu and deactivates keyboard monitoring.
###END###

*****************************************************************************/

static ERROR MENU_LostFocus(struct Menu *Self, APTR Void)
{
   OBJECTPTR keyboard;

   if ((Self->KeyboardID) AND (Self->KeyboardChannel IS TRUE)) {
      if (AccessObject(Self->KeyboardID, 4000, &keyboard) IS ERR_Okay) {
         UnsubscribeChannel(keyboard, Self->Head.UniqueID);
         ReleaseObject(keyboard);
      }
      else ObjectError(Self, ERH_ActionNotify, ERR_AccessObject);

      Self->KeyboardChannel = FALSE;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Menu: NewObject
*/

static ERROR MENU_NewObject(struct Menu *Self, APTR Void)
{
   if (NewObject(ID_FONT, NF_CHILD, &Self->Font, NULL) IS ERR_Okay) {
      if (NewObject(ID_XML, NF_CHILD, &Self->prvXML, NULL) IS ERR_Okay) {
         SetField(Self->Font, FID_Face, FT_STRING, "sans serif");
         Self->BreakHeight    = 6;
         Self->LeftMargin     = 5;
         Self->RightMargin    = 5;
         Self->TopMargin      = 3;
         Self->BottomMargin   = 3;
         Self->ImageGap       = 8;
         Self->ImageSize      = 16;
         Self->AutoExpand     = 200;
         Self->KeyGap         = 20;
         Self->VSpacing       = 4;
         Self->Highlight      = -1;
         Self->HighlightLM    = -1;
         Self->HighlightRM    = -1;
         Self->ExtensionGap   = 20;
         Self->Duration       = 250;
         Self->prvHighlight.Red   = 0;
         Self->prvHighlight.Green = 0;
         Self->prvHighlight.Blue  = 128;
         Self->prvFontHighlight.Red   = 255;
         Self->prvFontHighlight.Green = 255;
         Self->prvFontHighlight.Blue  = 255;
         Self->prvFontColour.Red   = 0;
         Self->prvFontColour.Green = 0;
         Self->prvFontColour.Blue  = 0;

         /*** Assume that the menu is in english ***/

         Self->Language[0] = 'e';
         Self->Language[1] = 'n';
         Self->Language[2] = 'g';
         Self->Language[3] = 0;

         StrCopy("lang", Self->LanguageDir, sizeof(Self->LanguageDir));

         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_NewObject, ERR_NewObject));
   }
   else return(ObjectError(Self, ERH_NewObject, ERR_NewObject));
}

/*****************************************************************************

###ACTION###
Name:  Refresh
Short: Refreshes a menu from its source file.
###END###

*****************************************************************************/

static ERROR MENU_Refresh(struct Menu *Self, APTR Void)
{
   OBJECTPTR object;
   LONG i, count;
   ERROR error;

   DPrintF("~Refresh()","[Menu:%d]", Self->Head.UniqueID);

   for (i=0; i < Self->TotalItems; i++) {
      if (Self->Items[i]->PictureID) ActionMsg(AC_Free, Self->Items[i]->PictureID, NULL);
      if (Self->Items[i]->MenuID)    ActionMsg(AC_Free, Self->Items[i]->MenuID, NULL);
      if (Self->Items[i]->Name)      FreeMemory(Self->Items[i]->Name);
      if (Self->Items[i]->Text)      FreeMemory(Self->Items[i]->Text);
      if (Self->Items[i]->Location)  FreeMemory(Self->Items[i]->Location);
      if (Self->Items[i]->ChildXML)  FreeMemory(Self->Items[i]->ChildXML);
      if (Self->Items[i]->Commands)  FreeMemory(Self->Items[i]->Commands);

      FreeMemory(Self->Items[i]);
      Self->Items[i] = NULL;
   }

   Self->TotalItems = 0;

   if (Self->DrawableID) {
      if (AccessObject(Self->DrawableID, 4000, &object) IS ERR_Okay) {
         UnsubscribeAction(object, NULL, GetUniqueID(Self));
         Action(AC_Free, object, NULL);
         ReleaseObject(object);
      }
      else ObjectError(Self, ERH_Refresh, ERR_AccessObject);
      Self->DrawableID = NULL;
   }

   /*** Destroy all child menus ***/

   DPrintF("~Refresh:","Destroying all child menus.");

   if ((count = TotalChildren(Self->Head.UniqueID))) {
      struct ChildEntry list[count];
      if (ListChildren(Self->Head.UniqueID, list, &count) IS ERR_Okay) {
         for (i=0; i < count; i++) {
            if (list[i].ClassID IS ID_MENU) ActionMsg(AC_Free, list[i].ObjectID, NULL);
         }
      }
   }

   StepBack();

   /*** Recreate the menu from scratch ***/

   DPrintF("Refresh:","Generating the new menu set.");

   Self->CurrentMenuID = NULL;

   if ((error = CreateMenu(Self)) IS ERR_Okay) {
      if (Self->DrawableID) {
         if (AccessObject(Self->DrawableID, 4000, &object) IS ERR_Okay) {
            ActionTags(MT_SetRenderCallback, object, Self, &DrawMenu);
            SubscribeActionTags(object, Self, AC_Show, AC_Hide, AC_Focus, AC_LostFocus, AC_UserClick, AC_UserMovement, TAGEND);
            ReleaseObject(object);
         }
         else return(ObjectError(Self, ERH_Refresh, ERR_AccessObject));
      }
      else return(ObjectError(Self, ERH_Refresh, ERR_FieldNotSet));

      StepBack();
      return(ERR_Okay);
   }
   else {
      StepBack();
      return(ObjectError(Self, ERH_Refresh, error));
   }
}

/*****************************************************************************

###ACTION###
Name:  SetUnlistedField
Short: Arguments can be passed to a menu configuration file as unlisted fields.
###END###

*****************************************************************************/

static ERROR MENU_SetUnlistedField(struct Menu *Self, struct acSetUnlistedField *Args)
{
   #define MAX_ARGS 40
   STRING str, arg;
   LONG i;

   /* Please note that it is okay to set zero-length string values
   ** (this has its uses in some scripts).
   */

   if ((!Args->Field) OR (!Args->Value)) return(ERR_Args);
   if (!Args->Field[0]) return(ERR_Args);

   if (!Self->LocalArgs) {
      if (AllocMemory(sizeof(STRING) * MAX_ARGS, Self->Head.MemFlags, (APTR *)&Self->LocalArgs, NULL) != ERR_Okay) {
         return(ERR_Memory);
      }
      Self->TotalArgs = 0;
   }
   else if (Self->TotalArgs >= MAX_ARGS) {
      DPrintF("@SetField:","[Menu:%d] Maximum number of arguments reached (%d) for \"%s\".", Self->Head.UniqueID, MAX_ARGS, Self->Location);
      return(ERR_Failed);
   }

   if (AllocMemory(StrLength(Args->Field) + StrLength(Args->Value) + 2,
                   MEM_STRING, (APTR *)&arg, NULL) IS ERR_Okay) {
      str = arg;
      for (i=0; Args->Field[i]; i++) *str++ = Args->Field[i];
      *str++ = 0;

      for (i=0; Args->Value[i]; i++) *str++ = Args->Value[i];
      *str = 0;

      /*** Argument insertion ***/

      for (i=0; i < Self->TotalArgs; i++) {
         str = Self->LocalArgs[i];
         if (StrCompare(arg, str, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            FreeMemory(Self->LocalArgs[i]);
            Self->LocalArgs[i] = arg;
            break;
         }
      }

      if (i IS Self->TotalArgs) {
         Self->LocalArgs[Self->TotalArgs] = arg;
         Self->TotalArgs++;
      }
   }
   else return(ERR_Memory);

   return(ERR_Okay);
}


/*****************************************************************************

###ACTION###
Name: Show
Short: Shows the menu.
###END###

*****************************************************************************/

static ERROR MENU_Show(struct Menu *Self, APTR Void)
{
   struct acMoveToPoint moveto;
   struct Render *drawable, *target, *relative;
   LONG ycoord, xcoord, absx, absy;

   if (AccessObject(Self->DrawableID, 4000, (OBJECTPTR *)&drawable) IS ERR_Okay) {

      if (Self->Flags & MNF_POPUP) {
         if (GetCursorPos(&xcoord, &ycoord) IS ERR_Okay) {
            OBJECTPTR parent;
            if (AccessObject(drawable->ParentID, 4000, &parent) IS ERR_Okay) {
               GetFields(parent, FID_AbsX|TLONG, &absx,
                                 FID_AbsY|TLONG, &absy,
                                 TAGEND);

               moveto.XCoord = xcoord - absx - 3;
               moveto.YCoord = ycoord - absy - 3;
               moveto.Flags = MTF_X|MTF_Y;
               Action(AC_MoveToPoint, drawable, &moveto);

               ReleaseObject(parent);
            }
         }
      }
      else if ((Self->RelativeID) AND (Self->TargetID)) {
         /* Correct the position of the menu according to the relative
         ** object that it is offset from.
         */

         if (AccessObject(Self->RelativeID, 4000, (OBJECTPTR *)&relative) IS ERR_Okay) {
            if (AccessObject(Self->TargetID, 4000, (OBJECTPTR *)&target) IS ERR_Okay) {

               GetFields(relative, FID_AbsX|TLONG, &xcoord,
                                   FID_AbsY|TLONG, &ycoord,
                                   TAGEND);

               moveto.XCoord = xcoord + Self->XCoord - target->XCoord;
               moveto.YCoord = ycoord + Self->YCoord - target->YCoord;
               moveto.Flags  = MTF_X|MTF_Y;

               if ((moveto.YCoord + drawable->Height) > target->Height) {
                  if (Self->ParentID) {
                     /* Use this code if we are a child menu */
                     moveto.YCoord = moveto.YCoord - drawable->Height + GetItemHeight(Self) + Self->VWhiteSpace;
                  }
                  else moveto.YCoord = moveto.YCoord - drawable->Height - Self->YCoord + Self->VWhiteSpace;
               }

               Action(AC_MoveToPoint, drawable, &moveto);

               ReleaseObject(target);
            }
            else ObjectError(Self, ERH_Activate, ERR_AccessObject);

            ReleaseObject(relative);
         }
         else ObjectError(Self, ERH_Activate, ERR_AccessObject);
      }

      acMoveToFront(drawable);
      acFocus(drawable);
      acShow(drawable);

      ReleaseObject(drawable);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Menu: Timer
*/

static ERROR MENU_Timer(struct Menu *Self, struct acTimer *Args)
{
   struct mtSetRenderOpacity setopacity;
   UBYTE unsubscribe;

   if (!Args) return(ObjectError(Self, ERH_Timer, ERR_Args));

   unsubscribe = TRUE;

   Self->FadeTimer += Args->TimeElapsed;
   if (Self->FadeTimer > Self->Duration) Self->FadeTimer = Self->Duration;

   if (Self->MouseOver > 0) {
      Self->MouseOver -= Args->TimeElapsed;

      if (Self->MouseOver <= 0) {
         Self->MouseOver = 0;
         if (Self->Highlight != -1) {
            if (Self->Items[Self->Highlight]->Flags & MENUITEM_EXTENSION) {
               ExecuteItem(Self, Self->Highlight, FALSE);
            }
         }
      }
      else unsubscribe = FALSE;
   }

   if (Self->prvFade IS MENUFADE_FADEIN) {
      setopacity.Value = (Self->FadeTimer * 100) / Self->Duration;
      if (setopacity.Value >= 100) setopacity.Value = 100;

      setopacity.Adjustment = 0;
      ActionMsg(MT_SetRenderOpacity, Self->DrawableID, &setopacity);

      if (setopacity.Value >= 100) Self->prvFade = NULL;
      else unsubscribe = FALSE;
   }
   else if (Self->prvFade IS MENUFADE_FADEOUT) {
      setopacity.Value = 100 - ((Self->FadeTimer * 100) / Self->Duration);
      if (setopacity.Value < 0) setopacity.Value = 0;

      setopacity.Adjustment = 0;
      ActionMsg(MT_SetRenderOpacity, Self->DrawableID, &setopacity);

      if (setopacity.Value < 1) {
         Self->prvFade = NULL;
         ActionMsg(AC_Hide, Self->DrawableID, NULL);
      }
      else unsubscribe = FALSE;
   }

   if (unsubscribe) {
      UnsubscribeTimer(Self->Head.UniqueID);
      Self->prvFade = 0;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   MenuFile
Short:  ...
Type:   STRING
Status: Set

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR SET_MenuFile(struct Menu *Self, STRING Value)
{
   OBJECTPTR xml;
   struct XMLTag **xmllist, *tag;
   ERROR error;

   if (!Value) return(ERR_Failed);

   DPrintF("~Set.MenuFile:","[Menu] %s", Value);

   error = ERR_Okay;
   if (CreateObject(ID_XML, NF_CHILD, &xml, NULL,
      FID_Location|TSTRING, Value,
      NULL) IS ERR_Okay) {

      if (GetFields(xml, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
         for (tag=xmllist[0]; tag; tag=tag->Next) {
            ParseXMLTag(Self, xml, tag);
         }
      }
      else error = ObjectError(Self, ERH_SetField, ERR_GetField);

      Action(AC_Free, xml, NULL);
   }
   else error = ObjectError(Self, ERH_SetField, ERR_CreateObject);

   StepBack();
   return(error);
}

/*****************************************************************************

*****************************************************************************/

static ERROR SET_FontColour(struct Menu *Self, STRING Colour)
{
   ReadColour(Colour, &Self->prvFontColour.Red, &Self->prvFontColour.Green, &Self->prvFontColour.Blue);
   return(ERR_Okay);
}

/*****************************************************************************

*****************************************************************************/

static ERROR SET_FontHighlight(struct Menu *Self, STRING Colour)
{
   ReadColour(Colour, &Self->prvFontHighlight.Red, &Self->prvFontHighlight.Green, &Self->prvFontHighlight.Blue);
   return(ERR_Okay);
}

/*****************************************************************************

*****************************************************************************/

static ERROR SET_Highlight(struct Menu *Self, STRING Colour)
{
   ReadColour(Colour, &Self->prvHighlight.Red, &Self->prvHighlight.Green, &Self->prvHighlight.Blue);
   return(ERR_Okay);
}

/*****************************************************************************

*****************************************************************************/

static ERROR SET_HighlightBorder(struct Menu *Self, STRING Colour)
{
   Self->Flags |= MNF_HIGHLIGHTBORDER;
   ReadColour(Colour, &Self->prvHighlightBorder.Red, &Self->prvHighlightBorder.Green, &Self->prvHighlightBorder.Blue);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:     Location
Synonyms: Src
Short:    Identifies the location of a menu configuration file to load.
Type:     STRING
Status:   Get/Set

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR GET_Location(struct Menu *Self, STRING *Value)
{
   if (Self->Location) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
}

static ERROR SET_Location(struct Menu *Self, STRING Value)
{
   LONG i;

   if (Self->Location) { FreeMemory(Self->Location); Self->Location = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->Location, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Location[i] = Value[i];
         Self->Location[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Node
Short:  The node that you wish to use for building a menu from its file source.
Type:   STRING
Status: Get/Set

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR GET_Node(struct Menu *Self, STRING *Value)
{
   if (Self->prvNode[0]) {
      *Value = Self->prvNode;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
}

static ERROR SET_Node(struct Menu *Self, STRING Value)
{
   LONG i;

   if (!Value) Self->prvNode[0] = 0;
   else {
      for (i=0; Value[i] AND (i < LEN_NODE-1); i++) {
         Self->prvNode[i] = Value[i];
      }
      Self->prvNode[i] = 0;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   SelectionObject
Short:  If the menu object should send user selections to another object, set this field.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>When the user selects a menu item, you can send the entered data to
the field of another object.  In order to do this, you need to point to that
object through the SelectionObject field and specify the destination field name
by setting the SelectionField string.  So long as both fields are set correctly,
the destination object will always reflect the last item selected from the
menu object.</p>

###SEE ALSO###
Field: SelectionField

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   SelectionField
Short:  This field is used in conjunction with the SelectionObject field.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>This field needs to be set in conjunction with the SelectionObject field
- please refer to it for more detail.</p>

###SEE ALSO###
Field: SelectionObject

###END###

*****************************************************************************/

static ERROR SET_SelectionField(struct Menu *Self, STRING Value)
{
   StrCopy(Value, Self->SelectionField, sizeof(Self->SelectionField));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  The pixel width of the menu.
Type:   LONG
Status: Read/Set

###DESCRIPTION###

###SEE ALSO###
Field: YCoord

###END###

*****************************************************************************/

static ERROR SET_Width(struct Menu *Self, LONG *Value)
{
   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      if (*Value > 0) {
         Self->Width = *Value;
         CalculateMenuSize(Self);
      }
   }
   else if (*Value >= 0) Self->Width = *Value;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal position of the menu.
Type:   LONG
Status: Set

###DESCRIPTION###

###SEE ALSO###
Field: YCoord

###END###

*****************************************************************************/

static ERROR SET_X(struct Menu *Self, LONG *Value)
{
   Self->XCoord = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical position of the menu.
Type:   LONG
Status: Set

###DESCRIPTION###

###SEE ALSO###
Field: XCoord

###END###

*****************************************************************************/

static ERROR SET_Y(struct Menu *Self, LONG *Value)
{
   Self->YCoord = *Value;
   return(ERR_Okay);
}

#include "functions.c"
