#ifndef SYSTEM_GLOBALBASE_H
#define SYSTEM_GLOBALBASE_H

/*
**  globalbase.h
**
**  (C) Copyright 1996-2004 Rocklyte Systems.
**      All Rights Reserved
*/

#ifndef KERNEL_H
#include <pandora/main.h>
#endif

/*
#ifdef __linux__
#include <pthread.h>
#undef NULL
#define NULL 0
#endif
*/

/*****************************************************************************
** Reserved Public Memory (RPM) ID's.  NOTE: RPM's are only viable for memory
** blocks that have a fixed size, as you cannot reallocate a memory block
** using the same ID.
*/

#define RPM_SharedObjects  -1000  /* Shared object list (struct SharedObject *) */
#define RPM_X11            -1004  /* X11 global variables */
#define RPM_AlphaBlend     -1005  /* Alpha blending lookup table */
#define RPM_XWindowLookup  -1006  /* X Window <-> Render lookup table */
#define RPM_NucleusExports -1007  /* Nucleus function exports */
#define RPM_FocusList      -1008  /* The Render class uses this to store the focus hierarchy */
#define RPM_RenderList     -1009  /* Manages render objects.  Only used by the Pointer and Render classes */

#ifdef __system__

/*****************************************************************************
** This structure is used for internally timed broadcasting.
*/

struct TimerBroadcast {
   OBJECTID SubscriberID;   /* The object that is subscribed */
   LONG     MilliSeconds;   /* The amount of milliseconds to wait at each interval */
   LONG     Count;          /* The amount of time currently left on the clock (resets at zero) */
   LARGE    Ticker;         /* Internal ticker for calculating elapsed times */
   LARGE    InitialMSec;    /* The time value recorded when the timer subscription was first activated */
};

/*****************************************************************************
** Shared object management.
*/

struct SharedObjectHeader {
   LONG Offset;          /* Offset of the main array - sizeof(struct PublicObjectHeader) */
   LONG NextEntry;       /* Next available entry within the array */
   LONG ArraySize;       /* Actual size of the array */
};

struct SharedObject {
   OBJECTID  ObjectID;          /* The object's ID */
   OBJECTID  OwnerID;           /* The owner of the object */
   MEMORYID  MessageMID;        /* If the object is private, this field refers to the Task MessageMID that owns it */
   OBJECTPTR Address;           /* Pointer the object address (if in private memory) */
   CLASSID   ClassID;           /* Class ID of the object */
   UBYTE     Name[MAX_NAMELEN]; /* Name of the object */
   UWORD     Flags;             /* NF flags */
   LONG      InstanceID;        /* Reference to the instance that this object is restricted to */
};

/*****************************************************************************
** Public memory management structures.
*/

struct PublicAddress {
   MEMORYID MemoryID;      /* Pandora memory ID */
   LONG     Size;          /* Size of the memory block */
   LONG     Offset;        /* Offset of the memory block within the page file */
   UBYTE    AccessCount;   /* Count of locks on this address */
   UBYTE    ExternalLock;  /* Incremented when a third party requires access during a lock */
   WORD     Flags;         /* Special MEM_ address flags */
   OBJECTID ObjectID;      /* Object that the address belongs to */
   OBJECTID ProcessLockID; /* If locked, this field refers to the ID of the process that locked the block */
   OBJECTID TaskID;        /* The Task that the block is tracked back to */
   LONG     InstanceID;    /* Reference to the instance that this memory block is restricted to */
   LONG     AccessTime;    /* The time at which the block was accessed */
   OBJECTID ContextID;     /* Context that locked the memory block (for debugging purposes only) */
   WORD     ActionID;      /* Action that locked the memory block (for debugging purposes only) */
   #ifdef _WIN32
      LONG ProcessID;      /* The process ID of the task that created the block as referred to by Offset */
      LONG Handle;         /* Memory handle, if block does not belong to the core memory pool */
   #endif
};

struct SharedControl {
   LONG PoolSize;          /* Amount of alllocated page space (starts at zero and expands) */
   LONG BlocksUsed;        /* Total amount of public memory blocks currently allocated */
   LONG MaxBlocks;         /* Maxmium amount of available blocks */
   LONG NextBlock;         /* Next empty position in the blocks table */
   LONG IDCounter;         /* ID counter */
   LONG PrivateIDCounter;  /* ID counter for private memory blocks */
   WORD PanicCounter;      /* Panic counter, increases when the public semaphore is inaccessible */
   WORD Locked;            /* Set to TRUE if a task is utilising shared memory management */
   LONG MessageIDCount;    /* Counter of message ID's */
   LONG ClassIDCount;      /* Counter of class ID's */
   LONG GlobalIDCount;     /* Counter for general ID's */
   LONG MagicKey;          /* This magic key is set to the semaphore key (used only as an indicator for initialisation) */
   LONG BlocksOffset;      /* Array of available shared memory pages */
   LONG SemaphoreOffset;   /* Offset to the semaphore control array */
   LONG TaskOffset;        /* Offset to the task control array */
   LONG MemoryOffset;      /* Offset to the public memory allocations */
   LONG GlobalInstance;    /* Set if a global instance is enabled */
   APTR FixedAddress;      /* Address used for fixed memory allocations */
   MEMORYID ObjectsMID;
   MEMORYID FieldNamesMID;
   MEMORYID FieldArrayMID;
   LONG ForbidProcess;     /* ID of the process that holds the current Forbid() */
   #ifdef __linux__
      //LONG ForbidLock;      /* Spinlock for forbid management */
      //pthread_mutex_t Mutex;
      //LONG ForbidQueue[4];
   #endif
};

#define MPF_LOCAL 0x0001       /* The page is owned by the task */
//#define MPF_POOL  0x0002       /* The page is part of a memory pool */

struct MemoryPage {
   MEMORYID MemoryID;          /* Represented memory ID */
   APTR     Address;           /* Map address */
   WORD     AccessCount;       /* Access count */
   WORD     Flags;             /* Special flags */
};

/*****************************************************************************
** Semaphore management structure.
*/

#define MAX_SEMAPHORES  40  /* Maximum number of semaphores that can be allocated in the system */

struct SemaphoreEntry {      /* The index of each semaphore in the array indicates their IDs */
   UBYTE    Name[16];        /* Name of the semaphore */
   LONG     InstanceID;      /* Reference to the instance that this semaphore is restricted to */
   LONG     BlockingProcess; /* Process ID of the blocker */
   WORD     Flags;           /* Status flags */
   WORD     BlockingValue;   /* Value used for blocking access */
   WORD     MaxValue;        /* Semaphore maximum value */
   WORD     Counter;         /* When the counter reaches zero, the semaphore is blocked */
   struct SemProcess {
      LONG ProcessID;
      UBYTE AllocCount;   /* Number of times that this process has allocated the semaphore with AllocSemaphore() */
      UBYTE BlockCount;   /* Count of blocking locks currently recorded */
      UBYTE AccessCount;  /* Count of access locks currently recorded */
      UBYTE BufferCount;  /* Buffered accesses (this value increases instead of AccessCount when the BlockCount is set) */
   } Processes[40];
   LONG     FIFO[10];       /* List of processes currently queued for access */
};

/*****************************************************************************
** Private memory management structures.
*/

struct PrivateAddress {
   APTR Address;
   MEMORYID MemoryID;
   OBJECTID ObjectID;
   LONG Size;
   WORD Flags;
   WORD AccessCount;
};

/****************************************************************************/

struct FieldEntry {
   STRING Name;
   LONG FieldID;
};

struct ResolveFieldArray {
   LONG TotalStrings;  /* The total number of strings currently listed in the field buffer */
   LONG BufferSize;    /* The byte size of the FieldNames memory buffer */
   LONG Offset;        /* The next byte offset to use for any new field name */
   LONG Index[4096];   /* list... */
};

/*****************************************************************************
** Message structure and internal ID's for standard Task-to-Task messages.
*/

#define SIZE_MSGBUFFER 16384

struct TaskMessage {
   LONG UniqueID;   /* Unique identifier for this particular message */
   LONG Type;       /* Message type ID */
   LONG DataSize;   /* Size of the data (does not include the size of the TaskMessage structure) */
   LONG NextMsg;    /* Offset to the next message */
   LONG Time;
   /*** Data follows ***/
};

struct MessageHeader {
   LONG NextEntry;     /* Byte offset for the next message to be stored */
   WORD Count;         /* Count of messages stored in the buffer */
   WORD TaskIndex;     /* Process that owns this message queue (refers to an index in the Task array) */
   LONG CompressReset; /* Manages message queue compression */
   BYTE Buffer[SIZE_MSGBUFFER + sizeof(struct TaskMessage)];
};

struct ActionMessage {
   WORD ActionID;            /* ID of the action or method to execute */
   WORD SendArgs;            /* */
   WORD ReturnResult;        /* Set to TRUE if a result is required */
   WORD Unused;              /* Unused variable */
   MEMORYID ReturnMessage;   /* If ReturnResult is TRUE, the message queue for the result message must be set here */
   OBJECTID ObjectID;        /* The object that is to receive the action */
   ERROR Error;              /* If a result is required, the action's error code is returned here by MSGID_ACTIONRESULT */
   LONG  Time;
   /*** Action arguments follow this structure in a buffer ***/
};

#endif

/*****************************************************************************
** X11 Variables
*/

#ifdef __xwindows__
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#undef NULL
#define NULL 0

struct WindowTable {
   LONG Size;
   LONG NextRecord;
};

struct WindowArray {
   Window Window;
   OBJECTID RenderID;
   OBJECTID ParentID;
};

#endif

#endif /* SYSTEM_GLOBALBASE_H */
